@@ -1,5 +1,7 @@
+from contextlib import contextmanager
 from sconce import monitors, schedules
-from sconce.exceptions import StopTrainingError
+from sconce.custom_exceptions import StopTrainingError
+
 
 import copy
 import numpy as np
@@ -7,6 +9,17 @@ import tempfile
 import torch
 
 
+@contextmanager
+def freezing(parameter_groups):
+    for pg in parameter_groups:
+        pg.freeze()
+    try:
+        yield parameter_groups
+    finally:
+        for pg in parameter_groups:
+            pg.unfreeze()
+
+
 class Trainer:
     """
     A Class that is used to train sconce models.
@@ -14,9 +27,9 @@ class Trainer:
     Keyword Arguments:
         model (:py:class:`~sconce.models.base.Model`): the sconce model to be trained.  See :py:mod:`sconce.models`
             for examples.
-        training_data_generator (:py:class:`~sconce.data_generators.base.DataGenerator`): yields training `inputs` and
+        training_feed (:py:class:`~sconce.data_feed.base.DataFeed`): used during training to provide `inputs` and
             `targets`.
-        test_data_generator (:py:class:`~sconce.data_generators.base.DataGenerator`): yields test `inputs` and
+        validation_feed (:py:class:`~sconce.data_feed.base.DataFeed`) used during validation to provide `inputs` and
             `targets`.  These are never used for back-propagation.
         monitor (:py:class:`~sconce.monitors.base.Monitor`, optional): the sconce monitor that records data during
             training.  This data can be sent to external systems during training or kept until training completes
@@ -24,25 +37,25 @@ class Trainer:
             :py:class:`~sconce.monitors.stdout_monitor.StdoutMonitor` and a
             :py:class:`~sconce.monitors.dataframe_monitor.DataframeMonitor` will be created for you and used.
     """
-    def __init__(self, *, model, test_data_generator, training_data_generator, monitor=None):
+    def __init__(self, *, model, validation_feed=None, training_feed=None, monitor=None):
         self.model = model
-        self.test_data_generator = test_data_generator
-        self.training_data_generator = training_data_generator
+
+        self.validation_feed = validation_feed
+        self.training_feed = training_feed
 
         if monitor is None:
-            metric_names = {'training_loss': 'loss', 'test_loss': 'val_loss'}
+            metric_names = {'training_loss': 'loss', 'validation_loss': 'val_loss'}
             stdout_monitor = monitors.StdoutMonitor(metric_names=metric_names)
             monitor = monitors.DataframeMonitor() + stdout_monitor
         self.monitor = monitor
 
-        self.test_to_train_ratio = (len(test_data_generator) /
-                                    len(training_data_generator))
+        self.validation_to_train_ratio = (len(validation_feed) / len(training_feed))
 
         self.checkpoint_filename = None
         self._reset_cache()
 
     def _reset_cache(self):
-        self._cache_data_generator = None
+        self._cache_feed = None
         self._inputs = None
         self._targets = None
         self._outputs = None
@@ -93,9 +106,7 @@ class Trainer:
         """
         self.model.load_state_dict(torch.load(filename))
 
-    def train(self, *, num_epochs, monitor=None,
-            test_to_train_ratio=None,
-            batch_multiplier=1):
+    def train(self, *args, **kwargs):
         """
         Train the model for a given number of epochs.
 
@@ -103,8 +114,8 @@ class Trainer:
             num_epochs (float): the number of epochs to train the model for.
             monitor (:py:class:`~sconce.monitors.base.Monitor`, optional): a monitor to use for this training session.
                 If ``None``, then self.monitor will be used.
-            test_to_train_ratio (float, optional): [0.0, 1.0] determines how often (relative to training samples) that
-                test samples are run through the model during training.  If ``None``, then the relative size of the
+            validation_to_train_ratio (float, optional): [0.0, 1.0] determines how often (relative to training samples)
+                that test samples are run through the model during training.  If ``None``, then the relative size of the
                 training and test datasets is used.  For example, for MNIST with 60,000 training samples and 10,000 test
                 samples, the value would be 1/6th.
             batch_multiplier (int, optional): [1, inf) determines how often parameter updates will occur during
@@ -112,6 +123,35 @@ class Trainer:
                 example, if the batch size were 100 and batch_multipler=10, the effective batch size would be 1,000, but
                 the memory usage would be for a batch size of 100.
 
+        Returns:
+            monitor (:py:class:`~sconce.monitors.base.Monitor`): the monitor used during training.
+        """
+        for monitor in self.get_training_iterator(*args, **kwargs):
+            continue
+        return monitor
+
+    def get_training_iterator(self, *, num_epochs, amp=False, monitor=None,
+            validation_to_train_ratio=None, batch_multiplier=1):
+        """
+        Get an iterator to train the model for a given number of epochs.
+
+        Arguments:
+            num_epochs (float): the number of epochs to train the model for.
+            amp (bool): Use automatic mixed precision training (defaults to False).
+            monitor (:py:class:`~sconce.monitors.base.Monitor`, optional): a monitor to use for this training session.
+                If ``None``, then self.monitor will be used.
+            validation_to_train_ratio (float, optional): [0.0, 1.0] determines how often (relative to training samples)
+                that test samples are run through the model during training.  If ``None``, then the relative size of the
+                training and test datasets is used.  For example, for MNIST with 60,000 training samples and 10,000 test
+                samples, the value would be 1/6th.
+            batch_multiplier (int, optional): [1, inf) determines how often parameter updates will occur during
+                training.  If greater than 1, this simulates large batch sizes without increasing memory usage.  For
+                example, if the batch size were 100 and batch_multipler=10, the effective batch size would be 1,000, but
+                the memory usage would be for a batch size of 100.
+
+        Yields:
+            monitor (:py:class:`~sconce.monitors.base.Monitor`): the monitor used during training.
+
         Returns:
             monitor (:py:class:`~sconce.monitors.base.Monitor`): the monitor used during training.
         """
@@ -120,32 +160,26 @@ class Trainer:
 
         if monitor is None:
             monitor = self.monitor
-        if test_to_train_ratio is None:
-            test_to_train_ratio = self.test_to_train_ratio
+        if validation_to_train_ratio is None:
+            validation_to_train_ratio = self.validation_to_train_ratio
 
         num_steps = self.get_num_steps(num_epochs=num_epochs,
-                data_generator=self.training_data_generator,
+                feed=self.training_feed,
                 batch_multiplier=batch_multiplier)
 
-        return self._train(num_steps=num_steps,
-                monitor=monitor,
-                test_to_train_ratio=test_to_train_ratio,
-                batch_multiplier=batch_multiplier)
-
-    def get_num_steps(self, num_epochs, data_generator=None, batch_multiplier=1):
-        if data_generator is None:
-            data_generator = self.training_data_generator
-        num_samples = num_epochs * data_generator.num_samples
-        batch_size = data_generator.batch_size
-        effective_batch_size = batch_size * batch_multiplier
-        num_steps = int(num_samples / effective_batch_size)
-
-        if num_steps * effective_batch_size < num_samples:
-            return num_steps + 1
+        if amp:
+            scaler = torch.cuda.amp.GradScaler()
         else:
-            return num_steps
+            scaler = None
 
-    def _train(self, *, num_steps, monitor, test_to_train_ratio, batch_multiplier):
+        return self._get_training_iterator(num_steps=num_steps,
+                monitor=monitor,
+                validation_to_train_ratio=validation_to_train_ratio,
+                batch_multiplier=batch_multiplier,
+                scaler=scaler)
+
+    def _get_training_iterator(self, *, num_steps, monitor, validation_to_train_ratio,
+            batch_multiplier, scaler=None):
         self._reset_cache()
         monitor.start_session(num_steps)
         self.model.start_session(num_steps)
@@ -165,34 +199,73 @@ class Trainer:
                 optimizer.zero_grad()
 
             for i in range(1, batch_multiplier + 1):
-                inputs, targets = self.training_data_generator.next()
-                step_dict = self._do_step(inputs, targets, train=True)
+                inputs, targets = self.training_feed.next()
+
+                if scaler is not None:
+                    with torch.cuda.amp.autocast():
+                        step_dict = self._do_step(inputs, targets, train=True)
+                else:
+                    step_dict = self._do_step(inputs, targets, train=True)
 
                 loss = step_dict['loss'] / batch_multiplier
-                loss.backward()
+                if scaler is not None:
+                    scaler.scale(loss).backward()
+                else:
+                    loss.backward()
 
-                training_step_dict = {f'training_{k}': v
-                        for k, v in step_dict.items()}
+                training_step_dict = {}
+                for k, v in step_dict.items():
+                    # now that we've backpropagated the loss, no need to keep computation graph.
+                    if hasattr(v, 'detach_'):
+                        v.detach_()
+                    training_step_dict[f'training_{k}'] = v
 
                 iterations_since_test += 1
-                if (1 / iterations_since_test) <= test_to_train_ratio:
-                    test_step_dict = self._do_test_step()
+                if (1 / iterations_since_test) <= validation_to_train_ratio:
+                    if scaler is not None:
+                        with torch.cuda.amp.autocast():
+                            validation_step_dict = self._do_validation_step()
+                    else:
+                        validation_step_dict = self._do_validation_step()
                     iterations_since_test = 0
 
-                    current_state.update({**training_step_dict, **test_step_dict})
+                    current_state.update({**training_step_dict, **validation_step_dict})
                 else:
                     current_state.update(training_step_dict)
 
                 fraction = i / batch_multiplier
                 monitor.write(data=current_state, step=step - 1 + fraction)
 
-            for optimizer in self.model.get_optimizers():
-                optimizer.step()
+            if scaler is not None:
+                for optimizer in self.model.get_optimizers():
+                    scaler.step(optimizer)
+            else:
+                for optimizer in self.model.get_optimizers():
+                    optimizer.step()
+
+            if scaler is not None:
+                scaler.update()
+
+            yield monitor
 
         monitor.end_session()
 
         return monitor
 
+    def get_num_steps(self, num_epochs, feed=None, batch_multiplier=1):
+        if feed is None:
+            feed = self.training_feed
+
+        num_samples = num_epochs * feed.num_samples
+        batch_size = feed.batch_size
+        effective_batch_size = batch_size * batch_multiplier
+        num_steps = int(num_samples / effective_batch_size)
+
+        if num_steps * effective_batch_size < num_samples:
+            return num_steps + 1
+        else:
+            return num_steps
+
     def _set_current_state(self, desired_state):
         for key, value in desired_state.items():
             if hasattr(self, f'set_{key}'):
@@ -216,12 +289,16 @@ class Trainer:
         self.model.train(train)
         in_dict = {'inputs': inputs, 'targets': targets}
 
-        out_dict = self.model(**in_dict)
+        if train:
+            out_dict = self.model(**in_dict)
+        else:
+            with freezing(self.model.active_parameter_groups):
+                out_dict = self.model(**in_dict)
         return {**out_dict, **in_dict}
 
-    def _run_model_on_generator(self, data_generator,
+    def _run_model_on_feed(self, feed,
             cache_results=True):
-        if self._cache_data_generator is data_generator:
+        if feed is self._cache_feed:
             return {'inputs': self._inputs,
                     'targets': self._targets,
                     'outputs': self._outputs}
@@ -230,9 +307,9 @@ class Trainer:
         targets = []
         outputs = []
 
-        data_generator.reset()
-        for x in range(len(data_generator)):
-            i, t = data_generator.next()
+        feed.reset()
+        for x in range(len(feed)):
+            i, t = feed.next()
             out_dict = self._run_model(i, t, train=False)
 
             inputs.append(i.cpu().data.numpy())
@@ -244,7 +321,7 @@ class Trainer:
         outputs = np.concatenate(outputs)
 
         if cache_results:
-            self._cache_data_generator = data_generator
+            self._cache_feed = feed
             self._inputs = inputs
             self._targets = targets
             self._outputs = outputs
@@ -253,14 +330,14 @@ class Trainer:
                 'targets': targets,
                 'outputs': outputs}
 
-    def _do_test_step(self):
-        inputs, targets = self.test_data_generator.next()
+    def _do_validation_step(self):
+        inputs, targets = self.validation_feed.next()
         step_dict = self._do_step(inputs, targets, train=False)
-        return {f'test_{k}': v for k, v in step_dict.items()}
+        return {f'validation_{k}': v for k, v in step_dict.items()}
 
-    def test(self, *, monitor=None):
+    def validate(self, *, monitor=None):
         """
-        Run all samples of self.test_data_generator through the model in test (inference) mode.
+        Run all samples of self.validation_feed through the model in test (inference) mode.
 
         Arguments:
             monitor (:py:class:`~sconce.monitors.base.Monitor`, optional): the sconce monitor that records data during
@@ -272,15 +349,15 @@ class Trainer:
             monitor (:py:class:`~sconce.monitors.base.Monitor`): the monitor used during this testing.
         """
         if monitor is None:
-            metric_names = {'test_loss': 'loss'}
+            metric_names = {'validation_loss': 'val_loss'}
             stdout_monitor = monitors.StdoutMonitor(metric_names=metric_names)
             monitor = monitors.DataframeMonitor() + stdout_monitor
 
-        num_steps = len(self.test_data_generator)
+        num_steps = len(self.validation_feed)
         monitor.start_session(num_steps)
 
         for step in range(1, num_steps + 1):
-            step_data = self._do_test_step()
+            step_data = self._do_validation_step()
 
             monitor.write(data=step_data, step=step)
         monitor.end_session()
@@ -312,7 +389,8 @@ class Trainer:
             max_learning_rate=10,
             monitor=None,
             batch_multiplier=1,
-            stop_factor=10):
+            stop_factor=10,
+            amp=False):
         """
         Checkpoints a model, then runs a learning rate survey, before restoring the model back.
 
@@ -352,8 +430,9 @@ class Trainer:
 
         self.train(num_epochs=num_epochs,
                 monitor=monitor,
-                test_to_train_ratio=0,
-                batch_multiplier=batch_multiplier)
+                validation_to_train_ratio=0,
+                batch_multiplier=batch_multiplier,
+                amp=amp)
 
         self.model.load_state_dict(orig_model_state_dict)
         for group, orig_optimizer_state_dict in zip(active_groups, orig_optimizer_state_dicts):
