@@ -1,7 +1,7 @@
 # flake8: noqa
-from sconce.data_generators import ImageDataGenerator
+from sconce.data_feeds import SingleClassImageFeed
 from sconce.schedules import Cosine
-from sconce.trainers import ClassifierTrainer
+from sconce.trainers import SingleClassImageClassifierTrainer
 from sconce.models import BasicClassifier
 from torch import optim
 
@@ -20,19 +20,19 @@ class TestBasicClassifier(unittest.TestCase):
         filename = self._test_file('basic_classifier.yaml')
         model = BasicClassifier.new_from_yaml_filename(filename)
 
-        training_generator = ImageDataGenerator.from_torchvision()
-        test_generator = ImageDataGenerator.from_torchvision(train=False)
+        training_feed = SingleClassImageFeed.from_torchvision()
+        validation_feed = SingleClassImageFeed.from_torchvision(train=False)
 
         if torch.cuda.is_available():
             model.cuda()
-            training_generator.cuda()
-            test_generator.cuda()
+            training_feed.cuda()
+            validation_feed.cuda()
 
         model.set_optimizer(optim.SGD, lr=1e-4, momentum=0.9, weight_decay=1e-4)
 
-        trainer = ClassifierTrainer(model=model,
-            training_data_generator=training_generator,
-            test_data_generator=test_generator)
+        trainer = SingleClassImageClassifierTrainer(model=model,
+            training_feed=training_feed,
+            validation_feed=validation_feed)
 
         self.assertLess(trainer.get_classification_accuracy(), 0.2)
 
